/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.idl.remote;

import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildPrimitiveType;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getContainedType;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getPrimitiveType;

import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ComplexType;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.PointerOf;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType;
import org.objectweb.fractal.cecilia.adl.idl.ast.RecordDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType.PrimitiveTypeEnum;

/**
 * Utility class providing methods to generate serialization and deserialization
 * code.
 */
public final class CSerializationHelper {

  private static final String BYTE_ARRAY_TYPE   = "minus.bare.generic.api.ByteArray";
  private static final String SHORT_ARRAY_TYPE  = "minus.bare.generic.api.ShortArray";
  private static final String USHORT_ARRAY_TYPE = "minus.bare.generic.api.UShortArray";
  private static final String INT_ARRAY_TYPE    = "minus.bare.generic.api.IntArray";
  private static final String UINT_ARRAY_TYPE   = "minus.bare.generic.api.UIntArray";

  private CSerializationHelper() {
  }

  static String getArrayCType(final Type type) {
    if (type instanceof PointerOf
        && getContainedType((PointerOf) type) instanceof ComplexType) {
      final IDLDefinition pointedDefinition = ((ComplexType) getContainedType((PointerOf) type))
          .getIDLDefinition();
      final String name = ((RecordDefinition) pointedDefinition).getName();
      if (BYTE_ARRAY_TYPE.equals(name))
        return "minus_bare_generic_api_ByteArray";
      else if (SHORT_ARRAY_TYPE.equals(name))
        return "minus_bare_generic_api_ShortArray";
      else if (USHORT_ARRAY_TYPE.equals(name))
        return "minus_bare_generic_api_UShortArray";
      else if (INT_ARRAY_TYPE.equals(name))
        return "minus_bare_generic_api_IntArray";
      else if (UINT_ARRAY_TYPE.equals(name))
        return "minus_bare_generic_api_UIntArray";
    }
    return null;
  }

  static String getArrayElementCType(final Type type) {
    if (type instanceof PointerOf
        && getContainedType((PointerOf) type) instanceof ComplexType) {
      final IDLDefinition pointedDefinition = ((ComplexType) getContainedType((PointerOf) type))
          .getIDLDefinition();
      final String name = ((RecordDefinition) pointedDefinition).getName();
      if (BYTE_ARRAY_TYPE.equals(name))
        return PrimitiveTypeEnum.BYTE.getCType();
      else if (SHORT_ARRAY_TYPE.equals(name))
        return PrimitiveTypeEnum.SHORT.getCType();
      else if (USHORT_ARRAY_TYPE.equals(name))
        return PrimitiveTypeEnum.UNSIGNED_SHORT.getCType();
      else if (INT_ARRAY_TYPE.equals(name))
        return PrimitiveTypeEnum.INT.getCType();
      else if (UINT_ARRAY_TYPE.equals(name))
        return PrimitiveTypeEnum.UNSIGNED_INT.getCType();
    }
    return null;
  }

  private static boolean isPointerToArray(final Type type) {
    if (type instanceof PointerOf
        && getContainedType((PointerOf) type) instanceof ComplexType) {
      final IDLDefinition pointedDefinition = ((ComplexType) getContainedType((PointerOf) type))
          .getIDLDefinition();
      return pointedDefinition instanceof RecordDefinition
          && (getArrayCType(type) != null);
    } else {
      return false;
    }
  }

  /**
   * Generates the declaration code required to the serialization of the
   * variable whose name is <code>varName</code> and its type is
   * <code>varType</code>. The returned object will be passed as
   * <code>context</code> parameter to the call of the
   * {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   * method for the same variable.
   * 
   * @param varType the type of the variable to serialize.
   * @param varName the name of the variable to serialize.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param validator the name validator. Can be used to validate a local
   *            variable name.
   * @return an object that will be passed as <code>context</code> parameter
   *         to the call of the
   *         {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   *         method for the same variable.
   */
  public static Object writeSerializationDeclaration(final Type varType,
      final String varName, final CodeWriter cw,
      final VarNameValidator validator) {
    if (varType instanceof PrimitiveType) {
      return writeSerializationDeclaration(
          getPrimitiveType((PrimitiveType) varType), varName, cw, validator);
    }
    if (isPointerToArray(varType)) {
      return null;
    }
    throw new IllegalArgumentException("Invalid varType");
  }

  /**
   * Generates the declaration code required to the serialization of the
   * variable whose name is <code>varName</code> and its type is
   * <code>varType</code>. The returned object will be passed as
   * <code>context</code> parameter to the call of the
   * {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   * method for the same variable.
   * 
   * @param varType the primitive type of the variable to serialize.
   * @param varName the name of the variable to serialize.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param validator the name validator. Can be used to validate a local
   *            variable name.
   * @return an object that will be passed as <code>context</code> parameter
   *         to the call of the
   *         {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   *         method for the same variable.
   */
  public static Object writeSerializationDeclaration(
      final PrimitiveTypeEnum varType, final String varName,
      final CodeWriter cw, final VarNameValidator validator) {
    switch (varType) {
      case INT :
      case SHORT :
      case UNSIGNED_INT :
      case UNSIGNED_SHORT :
      case UNSIGNED_BYTE :
      case BYTE :
      case BOOLEAN :
        return null;
      case STRING :
        final String l = validator.validateName(varName + "_length");
        cw.append("unsigned short ").append(l).append(" = 0;").endl();
        return l;
      default :
        throw new IllegalArgumentException(
            "This primitive type is not yet supported: " + varType);
    }
  }

  /**
   * Generates the declaration code required to the deserialization of the
   * variable whoes name is <code>varName</code> and its type is
   * <code>varType</code>. The returned object will be passed as
   * <code>context</code> parameter to the call of the
   * {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   * method for the same variable.
   * 
   * @param varType the type of the variable to serialize.
   * @param varName the name of the variable to serialize.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param validator the name validator. Can be used to validate a local
   *            variable name.
   * @return an object that will be passed as <code>context</code> parameter
   *         to the call of the
   *         {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   *         method for the same variable.
   */
  public static Object writeDeserializationDeclaration(final Type varType,
      final String varName, final CodeWriter cw,
      final VarNameValidator validator) {
    if (varType instanceof PrimitiveType) {
      return writeDeserializationDeclaration(
          getPrimitiveType((PrimitiveType) varType), varName, cw, validator);
    }
    if (isPointerToArray(varType)) {
      return null;
    }
    throw new IllegalArgumentException("Invalid varType");
  }

  /**
   * Generates the declaration code required to the deserialization of the
   * variable whoes name is <code>varName</code> and its type is
   * <code>varType</code>. The returned object will be passed as
   * <code>context</code> parameter to the call of the
   * {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   * method for the same variable.
   * 
   * @param varType the primitive type of the variable to serialize.
   * @param varName the name of the variable to serialize.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param validator the name validator. Can be used to validate a local
   *            variable name.
   * @return an object that will be passed as <code>context</code> parameter
   *         to the call of the
   *         {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   *         method for the same variable.
   */
  public static Object writeDeserializationDeclaration(
      final PrimitiveTypeEnum varType, final String varName,
      final CodeWriter cw, final VarNameValidator validator) {
    switch (varType) {
      case BOOLEAN :
      case INT :
      case SHORT :
      case BYTE :
      case UNSIGNED_BYTE :
      case UNSIGNED_INT :
      case UNSIGNED_SHORT :
      case UNSIGNED_LONG :
      case STRING :
        return null;
      default :
        throw new IllegalArgumentException(
            "This primitive type is not yet supported: " + varType);
    }
  }

  /**
   * Generates the serialization code for the variable whoes name is
   * <code>varName</code> and its type is <code>varType</code>. The
   * variable is serialized in the buffer <code>buffName</code>.
   * 
   * @param varType the type of the variable to serialize.
   * @param varName the name of the variable to serialize.
   * @param buffName the name of the buffer into which the variable will be
   *            serialized.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param context the object returned by the call of the
   *            {@link #writeSerializationDeclaration(Type, String, CodeWriter, VarNameValidator) writeSerializationDeclaration}
   *            method for the given variable.
   * @param free true if pointed area must be freed
   */
  public static void writeSerialization(final Type varType,
      final String varName, final String buffName, final CodeWriter cw,
      final Object context, final boolean free) {
    if (varType instanceof PrimitiveType) {
      writeSerialization(getPrimitiveType((PrimitiveType) varType), varName,
          buffName, cw, context, free);
      return;
    }
    if (isPointerToArray(varType)) {
      final IDLDefinition pointedDefinition = ((ComplexType) getContainedType((PointerOf) varType))
          .getIDLDefinition();
      final String name = ((RecordDefinition) pointedDefinition).getName();
      if (SHORT_ARRAY_TYPE.equals(name) || USHORT_ARRAY_TYPE.equals(name)) {
        cw.append(buffName).append(" = (char *)(((int)").append(buffName)
            .append(" + (1<<1) - 1) & ~((1<<1)-1));").endl();
      } else if (INT_ARRAY_TYPE.equals(name) || UINT_ARRAY_TYPE.equals(name)) {
        cw.append(buffName).append(" = (char *)(((int)").append(buffName)
            .append(" + (1<<2) - 1) & ~((1<<2)-1));").endl();
      }

      /* Serialize the size of the array */
      writeSerialization(PrimitiveTypeEnum.UNSIGNED_INT, varName + "->size",
          buffName, cw, null, false);

      /* Serialize the array elements */
      cw.appendln("{");
      cw.appendln("int i;");

      /*
       * Align short and integer array in the message
       */

      final String arrayElementCType = getArrayElementCType(varType);
      cw.append(arrayElementCType).append(" *data_array = (").append(
          arrayElementCType).append(" *)").append(buffName).appendln(";");
      cw.append("for (i = 0; i < ").append(varName).append("->size; i++) {");
      cw.append("data_array[i] = ").append(varName).append("->array[i];")
          .endl();
      cw.appendln("}");
      cw.append("data_array += i;").endl();
      cw.append(buffName).append(" = (char *)data_array;").endl();
      cw.appendln("}");
      if (free) {
        cw.append("CALL(REQUIRED.allocator, free, (void *)").append(varName)
            .append("->array);").endl();
        cw.append("CALL(REQUIRED.allocator, free, (void *)").append(varName)
            .append(");").endl();
      }
      return;
    }
    throw new IllegalArgumentException("Invalid varType");
  }

  /**
   * Generates the serialization code for the variable whoes name is
   * <code>varName</code> and its type is <code>varType</code>. The
   * variable is serialized in the buffer <code>buffName</code>.
   * 
   * @param varType the primitive type of the variable to serialize.
   * @param varName the name of the variable to serialize.
   * @param buffName the name of the buffer into which the variable will be
   *            serialized.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param context the object returned by the call of the
   *            {@link #writeSerializationDeclaration(Type, String, CodeWriter, VarNameValidator) writeSerializationDeclaration}
   *            method for the given variable.
   * @param free true if pointed area must be freed
   */
  public static void writeSerialization(final PrimitiveTypeEnum varType,
      final String varName, final String buffName, final CodeWriter cw,
      final Object context, final boolean free) {
    int i = 0;
    switch (varType) {
      case INT :
        writeSerialization(PrimitiveTypeEnum.UNSIGNED_INT, "((unsigned int) "
            + varName + ")", buffName, cw, context, false);
        return;
      case SHORT :
        writeSerialization(PrimitiveTypeEnum.UNSIGNED_SHORT,
            "((unsigned short) " + varName + ")", buffName, cw, context, false);
        return;
      case UNSIGNED_INT :
        cw.append(buffName).append("[").append(i++).append("] = (char) ((")
            .append(varName).append(" >> 24) & 0xFF);").endl();
        cw.append(buffName).append("[").append(i++).append("] = (char) ((")
            .append(varName).append(" >> 16) & 0xFF);").endl();
      case UNSIGNED_SHORT :
        cw.append(buffName).append("[").append(i++).append("] = (char) ((")
            .append(varName).append(" >> 8) & 0xFF);").endl();
      case UNSIGNED_BYTE :
      case BYTE :
      case BOOLEAN :
        cw.append(buffName).append("[").append(i++).append("] = (char) (")
            .append(varName).append(" & 0xFF);").endl();
        break;
      case STRING :
        final String length = (String) context;
        cw.appendln("{");
        cw.appendln("unsigned short i;");
        writeSerialization(PrimitiveTypeEnum.UNSIGNED_SHORT, length, buffName,
            cw, null, false);
        cw.append("for (i = 0; i < ").append(length).append("; i++) {").endl();
        cw.append(buffName).append("[i] = ").append(varName).append("[i];")
            .endl();
        cw.appendln("}");
        cw.appendln("}");
        if (free) {
          cw.append("CALL(REQUIRED.allocator, free, (void *)").append(varName)
              .append(");").endl();
        }
        cw.append(buffName).append(" += ").append(length).append(";").endl();
        return;
      default :
        throw new IllegalArgumentException(
            "This primitive type is not yet supported: " + varType);
    }
    cw.append(buffName).append(" += ").append(i).append(";").endl();
  }

  /**
   * Generates the deserialization code for the variable whoes name is
   * <code>varName</code> and its type is <code>varType</code>. The
   * variable is deserialized from the buffer <code>buffName</code>.
   * 
   * @param varType the type of the variable to deserialize.
   * @param varName the name of the variable to deserialize.
   * @param buffName the name of the buffer from which the variable will be
   *            deserialized.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param context the object returned by the call of the
   *            {@link #writeDeserializationDeclaration(Type, String, CodeWriter, VarNameValidator) writeSerializationDeclaration}
   *            method for the given variable.
   * @param alloc true if pointed area must be allocated and false if they can
   *            be static
   */
  public static void writeDeserialization(final Type varType,
      final String varName, final String buffName, final CodeWriter cw,
      final Object context, final boolean alloc) {
    if (varType instanceof PrimitiveType) {
      writeDeserialization(getPrimitiveType((PrimitiveType) varType), varName,
          buffName, cw, context, alloc);
      return;
    }
    if (isPointerToArray(varType)) {
      final String arrayCType = getArrayCType(varType);
      /*
       * Align short and integer array in the message
       */
      final IDLDefinition pointedDefinition = ((ComplexType) getContainedType((PointerOf) varType))
          .getIDLDefinition();
      final String name = ((RecordDefinition) pointedDefinition).getName();
      if (SHORT_ARRAY_TYPE.equals(name) || USHORT_ARRAY_TYPE.equals(name)) {
        cw.append(buffName).append(" = (char *)(((int)").append(buffName)
            .append(" + (1<<1) - 1) & ~((1<<1)-1));").endl();
      } else if (INT_ARRAY_TYPE.equals(name) || UINT_ARRAY_TYPE.equals(name)) {
        cw.append(buffName).append(" = (char *)(((int)").append(buffName)
            .append(" + (1<<2) - 1) & ~((1<<2)-1));").endl();
      }

      if (alloc) {
        cw.append(varName).append(" = (").append(arrayCType).append(" *) ")
            .append("CALL(REQUIRED.allocator, alloc, sizeof(").append(
                arrayCType).append("));").endl();

        writeDeserialization(PrimitiveTypeEnum.UNSIGNED_INT, varName
            + "->size ", buffName, cw, null, false);
        cw.append(varName).append("->array = CALL(REQUIRED.allocator, alloc,")
            .append(varName).append("->size * sizeof(").append(
                getArrayElementCType(varType)).append("));").endl();

        cw.appendln("{");
        cw.appendln("int i;");
        cw.append("for (i = 0; i <").append(varName).append("->size; i++) {")
            .endl();
        cw.append(varName).append("->array[i] = ((").append(
            getArrayElementCType(varType)).append(" *) ").append(buffName)
            .append(")[i];").endl();
        cw.appendln("}");

        cw.appendln("}");
        cw.append(buffName).append(" += ").append(varName).append(
            "->size * sizeof(").append(getArrayElementCType(varType)).append(
            ");").endl();
      } else {

        cw.append(varName).append(" = (").append(arrayCType).append(" *) ")
            .append("CALL(REQUIRED.allocator, alloc, sizeof(").append(
                arrayCType).append("));").endl();

        writeDeserialization(PrimitiveTypeEnum.UNSIGNED_INT, varName
            + "->size ", buffName, cw, null, false);
        cw.append(varName).append("->array = (").append(
            getArrayElementCType(varType)).append(" *)").append(buffName)
            .append(";").endl();
        cw.append(buffName).append(" += ").append(varName).append(
            "->size * sizeof(").append(getArrayElementCType(varType)).append(
            ");").endl();
      }
      return;
    }
    throw new IllegalArgumentException("Invalid varType");
  }

  /**
   * Generates the deserialization cleaning code for the variable whose name is
   * <code>varName</code> and its type is <code>varType</code>. The
   * variable is freed if it has been allocated.
   * 
   * @param varType the type of the variable to deserialize.
   * @param varName the name of the variable to deserialize.
   * @param buffName the name of the buffer from which the variable will be
   *            deserialized.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param context the object returned by the call of the
   *            {@link #writeDeserializationDeclaration(Type, String, CodeWriter, VarNameValidator) writeSerializationDeclaration}
   *            method for the given variable.
   */
  public static void writeDeserializationClean(final Type varType,
      final String varName, final String buffName, final CodeWriter cw,
      final Object context) {
    if (varType instanceof PrimitiveType) {
      return;
    }
    if (isPointerToArray(varType)) {
      return;
    }
    throw new IllegalArgumentException("Invalid varType");
  }

  /**
   * Generates the deserialization code for the variable whoes name is
   * <code>varName</code> and its type is <code>varType</code>. The
   * variable is deserialized from the buffer <code>buffName</code>.
   * 
   * @param varType the primitive type of the variable to deserialize.
   * @param varName the name of the variable to deserialize.
   * @param buffName the name of the buffer from which the variable will be
   *            deserialized.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param context the object returned by the call of the
   *            {@link #writeDeserializationDeclaration(Type, String, CodeWriter, VarNameValidator) writeSerializationDeclaration}
   *            method for the given variable.
   * @param alloc true if pointed area must be allocated and false if they can
   *            be static
   */
  public static void writeDeserialization(final PrimitiveTypeEnum varType,
      final String varName, final String buffName, final CodeWriter cw,
      final Object context, final boolean alloc) {
    int i = 0;
    final String buff = "((unsigned char *)" + buffName + ")";
    switch (varType) {
      case UNSIGNED_INT :
      case INT :
        cw.append(varName).append(" = ((unsigned int) ").append(buff).append(
            "[").append(i++).append("]) << 24;").endl();
        cw.append(varName).append(" |= ((unsigned int) ").append(buff).append(
            "[").append(i++).append("]) << 16;").endl();
        cw.append(varName).append(" |= ((unsigned int) ").append(buff).append(
            "[").append(i++).append("]) << 8;").endl();
        cw.append(varName).append(" |= (unsigned int) ").append(buff).append(
            "[").append(i++).append("];").endl();
        break;
      case UNSIGNED_SHORT :
      case SHORT :
        cw.append(varName).append(" = ((unsigned short) ").append(buff).append(
            "[").append(i++).append("]) << 8;").endl();
        cw.append(varName).append(" |= (unsigned short) ").append(buff).append(
            "[").append(i++).append("];").endl();
        break;
      case UNSIGNED_BYTE :
      case BYTE :
      case BOOLEAN :
        cw.append(varName).append(" = (unsigned char) ").append(buff).append(
            "[").append(i++).append("];").endl();
        break;
      case STRING :
        cw.appendln("{");
        cw.appendln("unsigned short length;");
        writeDeserialization(PrimitiveTypeEnum.UNSIGNED_SHORT, "length",
            buffName, cw, null, false);
        if (alloc == false) {
          cw.append(varName).append(" = (char *)").append(buff).append(";")
              .endl();
        } else {
          cw.appendln("unsigned short i;");
          cw.append(varName).append(
              " = CALL(REQUIRED.allocator, alloc, length + 1);").endl();
          cw.appendln("for (i = 0; i < length; i++) {");
          cw.append(varName).append("[i] = ").append(buff).append("[i];")
              .endl();
          cw.appendln("}");
          cw.append(varName).append("[length] = '\\0';").endl();
        }
        cw.append(buffName).append(" = (char *)(").append(buff).append(
            " + length);").endl();
        cw.appendln("}");
        return;
      default :
        throw new IllegalArgumentException(
            "This primitive type is not yet supported: " + varType);
    }
    cw.append(buffName).append(" = (char *)(").append(buff).append(" + ")
        .append(i).append(");").endl();
  }

  /**
   * Generates the code to determined the size of the variable whoes name is
   * <code>varName</code> and its type is <code>varType</code>. The size of
   * the variable is added to variable whoes name is <code>sizeofName</code>.
   * 
   * @param varType the type of the variable.
   * @param varName the name of the variable.
   * @param sizeofName the name of the variable incremeneted by the size of the
   *            variable <code>varName</code>.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param context the object returned by the call of the
   *            {@link #writeSerializationDeclaration(Type, String, CodeWriter, VarNameValidator) writeSerializationDeclaration}
   *            method for the given variable.
   */
  public static void writeSizeOf(final Type varType, final String varName,
      final String sizeofName, final CodeWriter cw, final Object context) {
    if (varType instanceof PrimitiveType) {
      writeSizeOf(getPrimitiveType((PrimitiveType) varType), varName,
          sizeofName, cw, context);
      return;
    }
    if (isPointerToArray(varType)) {
      /*
       * Short and Integer array will be aligned in the message, take into
       * account this alignment in the message size
       */
      final IDLDefinition pointedDefinition = ((ComplexType) getContainedType((PointerOf) varType))
          .getIDLDefinition();
      final String name = ((RecordDefinition) pointedDefinition).getName();
      if (SHORT_ARRAY_TYPE.equals(name) || USHORT_ARRAY_TYPE.equals(name)) {
        cw.append(sizeofName).append(" = (").append(sizeofName).append(
            " + (1<<1) - 1) & ~((1<<1)-1);");
      } else if (INT_ARRAY_TYPE.equals(name) || UINT_ARRAY_TYPE.equals(name)) {
        cw.append(sizeofName).append(" = (").append(sizeofName).append(
            " + (1<<2) - 1) & ~((1<<2)-1);");
      }

      cw.append(sizeofName).append(" += sizeof(").append(
          PrimitiveTypeEnum.UNSIGNED_INT.getCType()).append(") + ").append(
          varName).append("->size * sizeof(").append(
          getArrayElementCType(varType)).append(");").endl();
      return;
    }
    throw new IllegalArgumentException("Invalid varType");
  }

  /**
   * Generates the code to determined the size of the variable whoes name is
   * <code>varName</code> and its type is <code>varType</code>. The size of
   * the variable is added to variable whoes name is <code>sizeofName</code>.
   * 
   * @param varType the primitive type of the variable.
   * @param varName the name of the variable.
   * @param sizeofName the name of the variable incremeneted by the size of the
   *            variable <code>varName</code>.
   * @param cw the {@link CodeWriter} into which the generated code is appended.
   * @param context the object returned by the call of the
   *            {@link #writeSerializationDeclaration(Type, String, CodeWriter, VarNameValidator) writeSerializationDeclaration}
   *            method for the given variable.
   */
  public static void writeSizeOf(final PrimitiveTypeEnum varType,
      final String varName, final String sizeofName, final CodeWriter cw,
      final Object context) {
    switch (varType) {
      case INT :
      case SHORT :
      case UNSIGNED_INT :
      case UNSIGNED_SHORT :
      case UNSIGNED_BYTE :
      case BYTE :
      case BOOLEAN :
        cw.append(sizeofName).append(" += sizeof(").append(
            buildPrimitiveType(varType)).append(");").endl();
        break;
      case STRING :
        final String length = (String) context;
        cw.append("while (").append(varName).append("[").append(length).append(
            "++] != '\\0');").endl();
        cw.append(sizeofName).append(" += 2 + ").append(length).append(";")
            .endl();
        break;
      default :
        throw new IllegalArgumentException(
            "This primitive type is not yet supported: " + varType);
    }
  }

  /**
   * Returns <code>true</code> if the
   * {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   * and
   * {@link #writeDeserialization(Type, String, String, CodeWriter, Object, boolean) writeDeserialization}
   * can handle the given type.
   * 
   * @param type the type to test.
   * @return <code>true</code> if the
   *         {@link #writeSerialization(Type, String, String, CodeWriter, Object, boolean) writeSerialization}
   *         and
   *         {@link #writeDeserialization(Type, String, String, CodeWriter, Object, boolean) writeDeserialization}
   *         can handle the given type.
   */
  public static boolean isSerializable(final Type type) {
    if (type instanceof PrimitiveType) {
      switch (getPrimitiveType((PrimitiveType) type)) {
        case VOID :
        case BYTE :
        case SHORT :
        case INT :
        case UNSIGNED_BYTE :
        case UNSIGNED_SHORT :
        case UNSIGNED_INT :
        case STRING :
        case BOOLEAN :
          return true;
        default :
          return false;
      }
    }
    if (isPointerToArray(type)) {
      return true;
    }
    return false;
  }
}
