/***
 * Think ADL Parser
 * Copyright (C) 2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.interfaces;

import static org.objectweb.fractal.adl.NodeUtil.castNodeError;
import static org.objectweb.fractal.cecilia.adl.implementations.ImplementationDecorationUtil.getCode;

import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.CompilerError;
import org.objectweb.fractal.adl.ComponentVisitor;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.error.GenericErrors;
import org.objectweb.fractal.adl.implementations.Implementation;
import org.objectweb.fractal.adl.implementations.ImplementationContainer;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.cecilia.adl.AbstractDefinitionTask;
import org.objectweb.fractal.cecilia.adl.SourceCodeProvider;
import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.cecilia.adl.idl.InterfaceVisitor;
import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.FieldContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.MethodContainer;
import org.objectweb.fractal.task.core.AbstractTaskFactoryUser;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

/**
 * Abstract visitor component that creates a task that provides the
 * implementation source code of an interface adapter component.
 */
public abstract class AbstractInterfaceAdapterImplementationVisitor
    extends
      AbstractTaskFactoryUser implements ComponentVisitor {

  // ---------------------------------------------------------------------------
  // Implementation of the ComponentVisitor interface
  // ---------------------------------------------------------------------------

  public Component visit(final List<Node> path, final ComponentContainer node,
      final Map<Object, Object> context) throws ADLException, TaskException {
    final Implementation impl = castNodeError(node,
        ImplementationContainer.class).getImplementation();
    if (impl == null) {
      throw new CompilerError(GenericErrors.INTERNAL_ERROR,
          "This visitor is only applicable for primitive component.");
    }
    return createImplementationTask(node, impl);
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  /**
   * Creates the implementation task component.
   */
  protected Component createImplementationTask(final ComponentContainer node,
      final Implementation impl) throws TaskException {
    return taskFactoryItf.newPrimitiveTask(
        createImplementationTask((InterfaceDefinition) getCode(impl)), node);
  }

  /**
   * Returns a concrete implementation task object.
   */
  protected abstract AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf);

  // ---------------------------------------------------------------------------
  // Task classes
  // ---------------------------------------------------------------------------

  /**
   * Abstract task that generate implementation code for an interface adapter
   * component.
   */
  @TaskParameters("componentNode")
  @ServerInterfaces({@ServerInterface(name = "implementation-source-provider", signature = SourceCodeProvider.class, record = "role:implementationSourceCode, id:%", parameters = "componentNode")})
  public static abstract class AbstractInterfaceAdapterImplmentationTask
      extends
        AbstractDefinitionTask implements InterfaceVisitor {

    protected final InterfaceDefinition itf;
    protected CodeWriter                cw;

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public AbstractInterfaceAdapterImplmentationTask(
        final InterfaceDefinition itf) {
      this.itf = itf;
    }

    // -------------------------------------------------------------------------
    // Implementation of processSourceCode abstract method
    // -------------------------------------------------------------------------

    @Override
    protected String processSourceCode() throws Exception {
      cw = new CodeWriter(getClass().getName());
      visit(itf);
      return cw.toString();
    }

    public void visit(final IDLDefinition idlDefinition) throws Exception {
      final InterfaceDefinition itf = (InterfaceDefinition) idlDefinition;

      enterInterface(itf);

      for (final Field field : ((FieldContainer) itf).getFields())
        visitField(field);

      for (final Method method : ((MethodContainer) itf).getMethods())
        visitMethod(method);

      leaveInterface(itf);
    }
  }
}
