/***
 * Think ADL Parser
 * Copyright (C) 2005 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.cbst.eventProxy;

import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildComplexOrPrimitiveType;
import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildMethodDefinition;
import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildParameter;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getContainedType;

import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ArrayOf;
import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractInterfaceAdapterImplementationVisitor;

/**
 * Generates the implementation of a XStream event proxy component.
 */
public class ImplementationVisitor
    extends
      AbstractInterfaceAdapterImplementationVisitor {

  @Override
  protected AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf) {
    return new ImplementationTask(itf);
  }

  public static class ImplementationTask
      extends
        AbstractInterfaceAdapterImplmentationTask {

    private final String x       = new String(); // File
    // skeleton

    protected CodeWriter meths_cw;
    protected CodeWriter params_cw;
    protected CodeWriter args_cw;
    protected CodeWriter methCode_cw;
    protected CodeWriter execCode_cw;

    private String       ItfName = new String();

    private static enum ParseArrayMode {
      TYPE, SIZEOF
    };

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public ImplementationTask(final InterfaceDefinition itf) {
      super(itf);
    }

    private static String buildSizeableArray(Type type,
        final ParseArrayMode What) {
      final StringBuffer result = new StringBuffer();
      String arraySpec = "";

      while (type instanceof ArrayOf) {
        final String size = ((ArrayOf) type).getSize();
        if (size != null) {
          arraySpec = "[" + size + "]" + arraySpec;
        } else {
          arraySpec = "[]" + arraySpec;
        }
        type = getContainedType((ArrayOf) type);
      }
      if (What == ParseArrayMode.TYPE)
        result.append(buildComplexOrPrimitiveType(type));
      if (What == ParseArrayMode.SIZEOF) result.append(arraySpec);
      return result.toString();
    }

    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      meths_cw = new CodeWriter();
      params_cw = new CodeWriter();
      args_cw = new CodeWriter();
      methCode_cw = new CodeWriter();
      execCode_cw = new CodeWriter();

      ItfName = (itf.getName().replace('.', '_'));

      meths_cw.append("typedef enum {");

      params_cw.appendln("typedef struct {");
      params_cw.appendln(ItfName + "_meths_t Methods;");
      params_cw.appendln("union {");

      methCode_cw.appendln("void METHOD(Local, Registering) (void* _this)");
      methCode_cw.appendln("{");
      methCode_cw.appendln("if (DATA.ParamCount == ARRAY_SIZE)");
      methCode_cw.appendln("{");
      methCode_cw
          .appendln("printf (\"ERROR EvtProxy': Local recording overflow\");");
      methCode_cw.appendln("}");
      methCode_cw.appendln("CALL(REQUIRED.reg, RegisterEvent,");
      methCode_cw
          .appendln("GET_MY_INTERFACE(exec), (void*) DATA.ParamPushIdx);");
      methCode_cw.appendln("DATA.ParamCount++;");
      methCode_cw
          .appendln("if (++DATA.ParamPushIdx == ARRAY_SIZE) DATA.ParamPushIdx = 0;");
      methCode_cw.appendln("}");
      methCode_cw.appendln("//1) Receive client(/skeleton) call");
      methCode_cw.appendln("//2) Locally store parameters");

      execCode_cw
          .appendln("void METHOD(exec, ExecuteEvent)(void* _this, void* id)");
      execCode_cw.appendln("{");
      execCode_cw
          .appendln("// Retrieve function and parameters from structure pointed by id!");
      execCode_cw.appendln("switch (DATA.Param[(int)id].Methods)");
      execCode_cw.appendln("{");
    }

    public void leaveInterface(final InterfaceDefinition itf) throws Exception {
      meths_cw.appendln("} " + ItfName.toString() + "_meths_t;");

      params_cw.appendln("};");
      params_cw.appendln("} " + ItfName.toString() + "_parms_t," + " *p"
          + ItfName.toString() + "_parms_t;");

      methCode_cw
          .appendln("//4) Retrieve parameters when fired by Event Engine (execute protocol)");
      methCode_cw.appendln("//5) run server's service");

      execCode_cw.appendln("}");
      execCode_cw.appendln("DATA.ParamCount--;");
      execCode_cw.appendln("}");

      cw = new CodeWriter();
      cw.endl();

      cw.append("#include <stdio.h>").endl();
      cw.append(meths_cw.toString()).endl();
      cw.append(args_cw.toString()).endl();
      cw.append(params_cw.toString()).endl();

      // <<< Data Structures
      cw.appendln("// Able to handle  'ARRAY_SIZE' calls. FixMe: FIFO");
      cw.appendln("enum { ARRAY_SIZE=64 };");
      cw.appendln("DECLARE_DATA {");
      cw.appendln("unsigned int ParamPushIdx;");
      cw.appendln("unsigned int ParamPopIdx;");
      cw.appendln("unsigned int ParamCount;");
      cw.appendln("  " + ItfName + "_parms_t Param[ARRAY_SIZE];");
      cw.appendln("};");
      cw.appendln("#include <cecilia.h>");
      cw.endl();
      cw
          .appendln("//3) Post the call to Event Engine using registering protocol(Local meth)");

      cw.append(methCode_cw.toString()).endl();
      cw.append(execCode_cw.toString()).endl();

      cw.appendln("jint METHOD(lifecycle_controller, startFc)(void* _this)");
      cw.appendln("{");
      cw.appendln("DATA.ParamPushIdx = 0;");
      cw.appendln("DATA.ParamCount   = 0;");
      cw.appendln("return fractal_api_ErrorConst_OK;");
      cw.appendln("}");
      cw.endl();
      cw.appendln("jint METHOD(lifecycle_controller, stopFc)(void* _this)");
      cw.appendln("{");
      cw.appendln("return fractal_api_ErrorConst_OK;");
      cw.appendln("}");

      cw
          .appendln("jint METHOD(lifecycle_controller, getFcState)(void* _this) {");
      cw.appendln("return fractal_api_ErrorConst_OPERATION_NOT_SUPPORTED;");
      cw.appendln("}");

      cw.appendln(x);
    }

    public void visitField(final Field field) throws Exception {
    }

    public void visitMethod(final Method method) throws Exception {
      meths_cw.append(method.getName().toString() + "_ID, ");
      args_cw.append("typedef struct {" + "\n"); // args_t

      // Build ...METHOD(...)(...) signature
      methCode_cw.append(buildMethodDefinition(method, false, "METHOD(input,"
          + method.getName().toString() + ")"));

      methCode_cw.append('{' + "\n");
      methCode_cw
          .append(" // Allocate a structure containing functionID and paramaters!"
              + "\n");
      methCode_cw.append(" DATA.Param[DATA.ParamPushIdx].Methods="
          + method.getName().toString() + "_ID;" + "\n");
      for (final Parameter p : ((ParameterContainer) method).getParameters()) {
        methCode_cw.append(" DATA.Param[DATA.ParamPushIdx]."
            + method.getName().toString() + "_args." + p.getName() + "="
            + p.getName() + ";" + "\n");
      }
      methCode_cw.append(" CALLMINE(Local, Registering);" + "\n");
      methCode_cw.append('}' + "\n");

      execCode_cw.append("case " + method.getName().toString() + "_ID:" + "\n");
      execCode_cw.append("  CALL(REQUIRED.output,"
          + method.getName().toString());
      for (final Parameter p : ((ParameterContainer) method).getParameters()) {
        execCode_cw.append("," + "\n");
        execCode_cw.append("   DATA.Param[(int)id]."
            + method.getName().toString() + "_args." + p.getName());
      }
      execCode_cw.append(");" + "\n");
      execCode_cw.append("break;" + "\n");

      // args_t
      for (final Parameter p : ((ParameterContainer) method).getParameters()) {
        final Type type = getContainedType(p);

        if (type instanceof ArrayOf)
          // args_cw.append(buildSizeableArray(type, ParseArrayMode.TYPE) + "
          // (*"
          // + p.getName() + ")"
          // + buildSizeableArray(type, ParseArrayMode.SIZEOF) + ";\n");
          args_cw.append("void *" + p.getName() + ";\n");
        else
          args_cw.append(buildParameter(p) + ";\n");

      }
      args_cw.append("} " + method.getName().toString() + "_args_t;" + "\n");

      params_cw.append(method.getName().toString() + "_args_t "
          + method.getName().toString() + "_args" + ";\n");
    }
  }
}