/***
 * Think ADL Parser
 * Copyright (C) 2005 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.cbst.skel;

import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildComplexOrPrimitiveType;
import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildParameter;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getContainedType;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getPrimitiveType;

import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ArrayOf;
import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.PointerOf;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType.PrimitiveTypeEnum;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractInterfaceAdapterImplementationVisitor;

/**
 * Generates the implementation of a XStream skeleton component.
 */
public class ImplementationVisitor
    extends
      AbstractInterfaceAdapterImplementationVisitor {

  @Override
  protected AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf) {
    return new ImplementationTask(itf);
  }

  public static class ImplementationTask
      extends
        AbstractInterfaceAdapterImplmentationTask {

    private final String x       = new String(); // File
    // skeleton

    protected CodeWriter meths_cw;
    protected CodeWriter params_cw;
    protected CodeWriter args_cw;
    protected CodeWriter switchs_cw;

    private String       ItfName = new String();

    private static enum ParseArrayMode {
      TYPE, SIZEOF
    };

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public ImplementationTask(final InterfaceDefinition itf) {
      super(itf);
    }

    private static String buildSizeableArray(Type type,
        final ParseArrayMode What) {
      final StringBuffer result = new StringBuffer();
      String arraySpec = "";

      while (type instanceof ArrayOf) {
        final String size = ((ArrayOf) type).getSize();
        if (size != null) {
          arraySpec = "[" + size + "]" + arraySpec;
        } else {
          arraySpec = "[]" + arraySpec;
        }
        type = getContainedType((ArrayOf) type);
      }
      if (What == ParseArrayMode.TYPE)
        result.append(buildComplexOrPrimitiveType(type));
      if (What == ParseArrayMode.SIZEOF) result.append(arraySpec);
      return result.toString();
    }

    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      meths_cw = new CodeWriter();
      params_cw = new CodeWriter();
      args_cw = new CodeWriter();
      switchs_cw = new CodeWriter();

      ItfName = ("");
      ItfName = (itf.getName().replace('.', '_'));

      meths_cw.appendln("typedef enum {");

      params_cw.appendln("typedef struct {");
      params_cw.appendln(ItfName + "_meths_t Methods;");
      params_cw.appendln("union {");

      switchs_cw.appendln("switch (pPr->Methods)");
      switchs_cw.appendln("{");
    }

    public void leaveInterface(final InterfaceDefinition itf) throws Exception {

      cw = new CodeWriter();
      cw.endl();

      // >>> Data Structures
      meths_cw.appendln("} " + ItfName.toString() + "_meths_t;");

      params_cw.append("};");
      params_cw.append("} " + ItfName.toString() + "_parms_t," + " *p"
          + ItfName.toString() + "_parms_t;");

      switchs_cw.appendln("}");

      // <<< Data Structures
      cw.append(meths_cw.toString()).endl();
      cw.append(args_cw.toString()).endl();
      cw.append(params_cw.toString()).endl();

      cw.appendln("DECLARE_DATA {};");
      cw.appendln("#include <cecilia.h>");
      cw.endl();
      cw
          .appendln("int METHOD(push, push)(void *_this, Rminus_bare_generic_exportbind_api_Message *invokeMsg)");
      cw.appendln("{");
      cw.appendln("p" + ItfName + "_parms_t pPr;");
      cw.appendln("// Retrieve Param structure");
      cw
          .appendln("pPr = (p" + ItfName.toString()
              + "_parms_t)invokeMsg->data;");

      cw.appendln(switchs_cw.toString());

      cw
          .appendln("CALL(REQUIRED.message_manager, deleteMessageHeader, invokeMsg);");
      cw.appendln("CALL(REQUIRED.allocator, free, (jbyte *)pPr);");
      cw.appendln("return minus_bare_generic_exportbind_api_Push_E_OK;");
      cw.appendln("}");
      cw.endl();
    }

    public void visitField(final Field field) throws Exception {
    }

    /**
     * Returns the equivalent in C of the given return type.
     * 
     * @param type the type
     * @return the equivalent in C of the given return type.
     */
    public static String buildType(Type type) {
      final StringBuffer result = new StringBuffer();
      int nbPointer = 0;

      // A return type is an [ArrayOf](PointerOf)*(PrimitiveType|ComplexType)
      while (type instanceof ArrayOf) {
        nbPointer++;
        type = getContainedType((ArrayOf) type);
      }

      while (type instanceof PointerOf) {
        nbPointer++;
        type = getContainedType((PointerOf) type);
      }

      result.append(buildComplexOrPrimitiveType(type));

      for (int i = 0; i < nbPointer; i++)
        result.append('*');

      return result.toString();
    }

    // Memo: Skeleton
    public void visitMethod(final Method method) throws Exception {
      // methods_t
      meths_cw.append(method.getName().toString() + "_ID, ");
      // args_t
      args_cw.append("typedef struct {" + "\n");
      // switch code
      switchs_cw.append("case " + method.getName().toString() + "_ID:" + "\n");

      final int BlocksIdx = 1;
      Type type;
      for (final Parameter p : ((ParameterContainer) method).getParameters()) {
        type = getContainedType(p);
        boolean pointer = false;
        pointer |= ((type instanceof PrimitiveType) && (getPrimitiveType((PrimitiveType) type) == PrimitiveTypeEnum.STRING));
        pointer |= (type instanceof PointerOf);
        pointer |= (type instanceof ArrayOf);
        if (pointer) {
          switchs_cw.appendln("if(" + "pPr->" + method.getName() + "_args."
              + p.getName() + ")");
          switchs_cw.appendln("{");
          switchs_cw.appendln("pPr->" + method.getName() + "_args."
              + p.getName() + " = (" + buildType(type)
              + ")invokeMsg->next->data;");
          switchs_cw.appendln("invokeMsg = invokeMsg->next;");
          switchs_cw.appendln("}");
        }
        // args_t
        if (type instanceof ArrayOf)
          args_cw.append("void *" + p.getName() + ";\n");
        else
          args_cw.append(buildParameter(p) + ";\n");
      }
      switchs_cw.append("CALL(REQUIRED.itf," + method.getName().toString());
      for (final Parameter p : ((ParameterContainer) method).getParameters()) {
        switchs_cw
            .append(", pPr->" + method.getName() + "_args." + p.getName());
      }
      // switch code
      switchs_cw.appendln(");");
      switchs_cw.appendln("break;");

      // args_t
      args_cw.append("} " + method.getName().toString() + "_args_t;" + "\n");

      // params_t
      params_cw.append(method.getName().toString() + "_args_t "
          + method.getName().toString() + "_args" + ";\n");
    }
  }
}