/***
 
 * Think ADL Parser
 * Copyright (C) 2005 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.cbst.stub;

import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildComplexOrPrimitiveType;
import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildMethodDefinition;
import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildParameter;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getContainedType;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getPrimitiveType;

import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ArrayOf;
import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.PointerOf;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType.PrimitiveTypeEnum;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractInterfaceAdapterImplementationVisitor;

/**
 * Generates the implementation of a XStream stub component.
 */
public class ImplementationVisitor
    extends
      AbstractInterfaceAdapterImplementationVisitor {

  @Override
  protected AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf) {
    return new ImplementationTask(itf);
  }

  public static class ImplementationTask
      extends
        AbstractInterfaceAdapterImplmentationTask {

    protected CodeWriter meths_cw;
    protected CodeWriter params_cw;
    protected CodeWriter code_cw;
    protected CodeWriter args_cw;
    protected String     ItfName;

    private static enum ParseArrayMode {
      TYPE, SIZEOF
    };

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public ImplementationTask(final InterfaceDefinition itf) {
      super(itf);
    }

    private static String buildSizeableArray(Type type,
        final ParseArrayMode What) {
      final StringBuffer result = new StringBuffer();
      String arraySpec = "";

      while (type instanceof ArrayOf) {
        final String size = ((ArrayOf) type).getSize();
        if (size != null) {
          arraySpec = "[" + size + "]" + arraySpec;
        } else {
          arraySpec = "[]" + arraySpec;
        }
        type = getContainedType((ArrayOf) type);
      }
      if (What == ParseArrayMode.TYPE)
        result.append(buildComplexOrPrimitiveType(type));
      if (What == ParseArrayMode.SIZEOF) result.append(arraySpec);
      return result.toString();
    }

    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      meths_cw = new CodeWriter();
      params_cw = new CodeWriter();
      code_cw = new CodeWriter();
      args_cw = new CodeWriter();

      ItfName = (itf.getName().replace('.', '_'));

      meths_cw.endl();
      meths_cw.append("typedef enum {");

      params_cw.appendln("typedef struct {");
      params_cw.append(ItfName).append("_meths_t Methods;").endl();
      params_cw.appendln("union {");
    }

    public void leaveInterface(final InterfaceDefinition itf) throws Exception {
      meths_cw.append("NB_METHS} ").append(ItfName.toString()).append(
          "_meths_t;").endl();

      params_cw.appendln("};");
      params_cw.append("} ").append(ItfName.toString()).append("_parms_t, *p")
          .append(ItfName.toString()).append("_parms_t;").endl();

      cw = new CodeWriter();
      cw.appendln("DECLARE_DATA {};");
      cw.appendln("#include <cecilia.h>");
      cw.endl();
      cw.append(meths_cw.toString()).endl();
      cw.append(args_cw.toString()).endl();
      cw.append(params_cw.toString()).endl();
      cw.append(code_cw.toString());
    }

    public void visitField(final Field field) throws Exception {
    }

    // Memo: Stub
    public void visitMethod(final Method method) throws Exception {
      // methods_t
      meths_cw.append(method.getName().toString() + "_ID, ");
      // args_t
      args_cw.appendln("typedef struct {");

      // Build ...METHOD(...)(...) signature
      code_cw.appendln(buildMethodDefinition(method, false, "METHOD(itf,"
          + method.getName().toString() + ")"));

      // METHOD's code
      code_cw.appendln("{");

      code_cw
          .appendln("Rminus_bare_generic_exportbind_api_Message *msg, *firstMsg;");
      code_cw.endl();

      code_cw.appendln("p" + ItfName.toString() + "_parms_t pPr=" + "(p"
          + ItfName.toString()
          + "_parms_t)CALL(REQUIRED.allocator, alloc, sizeof("
          + ItfName.toString() + "_parms_t));" + "\n");
      code_cw.appendln("pPr->Methods = " + method.getName() + "_ID;");

      code_cw
          .appendln("firstMsg = CALL(REQUIRED.message_manager, createInitializedMessage, (jbyte *)pPr, sizeof("
              + ItfName.toString() + "_parms_t));");
      code_cw.appendln("msg = firstMsg;");
      code_cw.endl();

      Type type;
      for (final Parameter p : ((ParameterContainer) method).getParameters()) {
        type = getContainedType(p);

        if (type instanceof ArrayOf)
          code_cw.appendln("pPr->" + method.getName() + "_args." + p.getName()
              + " = " + p.getName() + ";" + "\n");
        else
          code_cw.appendln("pPr->" + method.getName() + "_args." + p.getName()
              + " = " + p.getName() + ";" + "\n");

        if ((type instanceof PrimitiveType)
            && (getPrimitiveType((PrimitiveType) type) == PrimitiveTypeEnum.STRING)) {
          code_cw.appendln("if(" + p.getName() + ")");
          code_cw.appendln("{");
          code_cw.appendln("int l=0; while (" + p.getName().toString()
              + "[l++]); // l=(strlen(pString)+1)");
          code_cw
              .appendln("msg->next = CALL(REQUIRED.message_manager, createInitializedMessage, (jbyte *)"
                  + p.getName() + ", l);");
          code_cw.appendln("msg = msg->next;");
          code_cw.appendln("}");
        }
        if (type instanceof PointerOf) {
          code_cw.appendln("if(" + p.getName() + ")");
          code_cw.appendln("{");
          code_cw
              .appendln("msg->next = CALL(REQUIRED.message_manager, createInitializedMessage, (jbyte *)"
                  + p.getName() + ", sizeof(*" + p.getName() + "));");
          code_cw.appendln("msg = msg->next;");
          code_cw.appendln("}");
        }
        if (type instanceof ArrayOf) {
          code_cw.appendln("if(" + p.getName() + ")");
          code_cw.appendln("{");
          code_cw
              .appendln("msg->next = CALL(REQUIRED.message_manager, createInitializedMessage, (jbyte *)"
                  + p.getName()
                  + ", sizeof("
                  + buildSizeableArray(type, ParseArrayMode.TYPE)
                  + buildSizeableArray(type, ParseArrayMode.SIZEOF) + "));");
          code_cw.appendln("msg = msg->next;");
          code_cw.appendln("}");
        }

        // args_t
        if (type instanceof ArrayOf)
          args_cw.appendln("void *" + p.getName() + ";");
        else
          args_cw.appendln(buildParameter(p) + ";");
      }
      args_cw.appendln("} " + method.getName().toString() + "_args_t;");

      params_cw.appendln(method.getName().toString() + "_args_t "
          + method.getName().toString() + "_args" + ";");

      code_cw.endl();
      code_cw.appendln("msg->next = 0;");
      code_cw.endl();
      code_cw.appendln("CALL(REQUIRED.push, push, firstMsg);");
      code_cw.appendln("}");
    }
  }
}