/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.rpc.skel;

import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeDeserialization;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeDeserializationClean;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeDeserializationDeclaration;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeSerialization;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeSerializationDeclaration;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeSizeOf;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getPrimitiveType;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.MethodContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType.PrimitiveTypeEnum;
import org.objectweb.fractal.cecilia.adl.idl.remote.VarNameValidator;
import org.objectweb.fractal.cecilia.adl.idl.util.CUtil;
import org.objectweb.fractal.cecilia.adl.idl.util.Util;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractInterfaceAdapterImplementationVisitor;

/**
 * Generates the implementation of a RPC stub component.
 */
public class SkeletonImplementationVisitor
    extends
      AbstractInterfaceAdapterImplementationVisitor {

  @Override
  protected AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf) {
    return new SkeletonImplementationTask(itf);
  }

  public static class SkeletonImplementationTask
      extends
        AbstractInterfaceAdapterImplmentationTask {

    private int              methodIdValue;
    private VarNameValidator localVars;
    private String           invokeMsg;
    private String           replyMsg;
    private String           replyMsgSize;
    private String           data;
    private String           methodId;
    private String           result;

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public SkeletonImplementationTask(final InterfaceDefinition itf) {
      super(itf);
    }

    /**
     * Builds the header part for a given interface definition.
     * 
     * @param itf the interface definition for which has to be built.
     * @throws Exception if any error occurs.
     */
    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      localVars = new VarNameValidator();

      // Enter name of every method parameters in VarNameValidator to ensure
      // that local variables will not clash.
      for (final Method method : ((MethodContainer) itf).getMethods()) {
        for (final Parameter p : ((ParameterContainer) method).getParameters()) {
          localVars.validateName(p.getName());
        }
      }

      methodIdValue = 0;
      invokeMsg = localVars.validateName("invokeMsg");
      replyMsg = localVars.validateName("replyMsg");
      replyMsgSize = localVars.validateName("replyMsgSize");
      data = localVars.validateName("data");
      methodId = localVars.validateName("methodId");
      result = localVars.validateName("result");

      final String componentCName = typeNameProviderItf.getCTypeName();
      cw.appendln("#include \"minus/bare/generic/exportbind/api/serialize.h\"");

      cw.append("#define REQUIRED ((struct ").append(componentCName).append(
          "_t *)_this)->type.imported ").endl();
      cw.append("#define NEWSTYLE_COMPONENT").endl();

      cw.append("  struct ").append(componentCName).append("_instancedata {};")
          .endl();
      cw.appendln("#include \"cecilia.h\"");
      cw.appendln("#include \"stdio.h\"");
      cw.endl();
      cw.append("Rminus_bare_generic_exportbind_api_Message *").append(
          componentCName).append("_invoke_invoke_method ").append(
          "(void *_this, Rminus_bare_generic_exportbind_api_Message *").append(
          invokeMsg).append(") {").endl();
      cw.append("Rminus_bare_generic_exportbind_api_Message *")
          .append(replyMsg).append(" = 0;").endl();
      cw.append("int ").append(replyMsgSize).append(" = 0;").endl();
      cw.append("char *").append(data).append(" = (char *)").append(invokeMsg)
          .append("->data;").endl();
      cw.append("char ").append(methodId).append(" = *(").append(data).append(
          "++);").endl();
      cw.append("switch (").append(methodId).append(") {").endl();
    }

    /**
     * Builds the footer part for a given interface definition.
     * 
     * @param itf the interface definition for which has to be built.
     * @throws Exception if any error occurs.
     */
    public void leaveInterface(final InterfaceDefinition itf) throws Exception {
      cw.append("default : printf(\"Invalid methodId %d\", ").append(methodId)
          .append(");").endl();
      cw.appendln("}");
      cw.append("return ").append(replyMsg).append(";").endl();
      cw.appendln("}");

      cw.append("#undef NEWSTYLE_COMPONENT").endl();
    }

    /**
     * Visits a field definition.
     * 
     * @param field the field definition to be used.
     * @throws Exception if any error occurs.
     */
    public void visitField(final Field field) throws Exception {
      // nothing todo
    }

    /**
     * Visits a method definition.
     * 
     * @param method the field definition to be used.
     * @throws Exception if any error occurs.
     */
    public void visitMethod(final Method method) throws Exception {
      final Type returnType = Util.getContainedType((method));
      final Parameter[] parameters = ((ParameterContainer) method)
          .getParameters();
      final boolean isVoidMethod = (returnType instanceof PrimitiveType)
          && getPrimitiveType((PrimitiveType) returnType) == PrimitiveTypeEnum.VOID;
      final Map<Parameter, Object> paramContext = new HashMap<Parameter, Object>();
      final Map<Parameter, Object> outParamContext = new HashMap<Parameter, Object>();
      Object resultContext = null;
      boolean hasOutParam = false;

      // determine if there is an Out parameter
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p) || Util.isInOutParameter(p))
          hasOutParam = true;
      }

      cw.append("case ").append(methodIdValue++).append(" :").endl();
      cw.appendln("{");

      // declare parameter variables
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p) || Util.isInOutParameter(p)) {
          final String qualifier = p.getQualifier();
          p.setQualifier(Parameter.IN);
          cw.append(CUtil.buildParameter(p)).append(";").endl();
          p.setQualifier(qualifier);
        } else {
          cw.append(CUtil.buildParameter(p)).append(";").endl();
        }
      }

      // declaration related to parameter deserialization
      for (final Parameter p : parameters) {
        final Object ctx = writeDeserializationDeclaration(Util
            .getContainedType(p), p.getName(), cw, localVars);
        paramContext.put(p, ctx);
      }

      if (!isVoidMethod) {
        // declare result variable.
        cw.append(CUtil.buildReturnType(returnType)).append(" ").append(result)
            .append(";").endl();
        // declaration related to the result value serialization
        resultContext = writeSerializationDeclaration(returnType, result, cw,
            localVars);
      }

      if (hasOutParam) {
        // declaration related to the out parameter serialization
        for (final Parameter p : parameters) {
          if (Util.isOutParameter(p) || Util.isInOutParameter(p)) {
            final Object ctx = writeSerializationDeclaration(Util
                .getContainedType(p), p.getName(), cw, localVars);
            outParamContext.put(p, ctx);
          }
        }
      }

      // deserialize parameters
      cw.append("// Deserialize parameters").endl();
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p)) continue;
        cw.append("// Deserialize \"" + p.getName() + "\"").endl();
        writeDeserialization(Util.getContainedType(p), p.getName(), data, cw,
            paramContext.get(p), false);
      }
      cw.endl();

      // invoke methods
      cw.append("// Invoke method").endl();
      if (!isVoidMethod) cw.append(result).append(" = ");
      cw.append("CALL(REQUIRED.itf, ").append(method.getName());
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p) || Util.isInOutParameter(p))
          cw.append(", &").append(p.getName());
        else
          cw.append(", ").append(p.getName());
      }
      cw.append(");").endl().endl();

      // clean deserialized parameters
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p)) continue;
        writeDeserializationClean(Util.getContainedType(p), p.getName(), data,
            cw, paramContext.get(p));
      }

      // create reply message
      if (!isVoidMethod || hasOutParam) {
        if (!isVoidMethod)
          writeSizeOf(returnType, result, replyMsgSize, cw, resultContext);

        for (final Parameter p : parameters) {
          if (Util.isOutParameter(p) || Util.isInOutParameter(p)) {
            writeSizeOf(Util.getContainedType(p), p.getName(), replyMsgSize,
                cw, outParamContext.get(p));
          }
        }
      }
      cw.append(replyMsg).append(
          " = CALL(REQUIRED.message_manager, createMessage, ").append(
          replyMsgSize).append(");").endl().endl();

      if (!isVoidMethod || hasOutParam) {
        cw.append("// Serialize return value, out and in/out parameters")
            .endl();
        cw.append("// Serialize return value").endl();
        cw.append(data).append(" = (char *)").append(replyMsg)
            .append("->data;").endl();
        if (!isVoidMethod)
          writeSerialization(returnType, result, data, cw, resultContext, true);
        for (final Parameter p : parameters) {
          if (Util.isOutParameter(p) || Util.isInOutParameter(p)) {
            cw.append("// Serialize \"" + p.getName() + "\"").endl();
            writeSerialization(Util.getContainedType(p), p.getName(), data, cw,
                outParamContext.get(p), true);
          }
        }
      }
      cw.appendln("break;");
      cw.appendln("}");
      cw.endl();
    }
  }
}
