/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.rpc.stub;

import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeDeserialization;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeDeserializationDeclaration;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeSerialization;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeSerializationDeclaration;
import static org.objectweb.fractal.cecilia.adl.idl.remote.CSerializationHelper.writeSizeOf;
import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildMethodDefinition;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getPrimitiveType;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.ast.PrimitiveType.PrimitiveTypeEnum;
import org.objectweb.fractal.cecilia.adl.idl.remote.VarNameValidator;
import org.objectweb.fractal.cecilia.adl.idl.util.CUtil;
import org.objectweb.fractal.cecilia.adl.idl.util.Util;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractInterfaceAdapterImplementationVisitor;

/**
 * Generates the implementation of a RPC stub component.
 */
public class StubImplementationVisitor
    extends
      AbstractInterfaceAdapterImplementationVisitor {

  @Override
  protected AbstractInterfaceAdapterImplmentationTask createImplementationTask(
      final InterfaceDefinition itf) {
    return new StubImplementationTask(itf);
  }

  public static class StubImplementationTask
      extends
        AbstractInterfaceAdapterImplmentationTask {

    private int methodId;

    /**
     * @param itf the {@link InterfaceDefinition} that is processed by this
     *            task.
     */
    public StubImplementationTask(final InterfaceDefinition itf) {
      super(itf);
    }

    /**
     * Builds the header part for a given interface definition.
     * 
     * @param itf the interface definition for which has to be built.
     * @throws Exception if any error occurs.
     */
    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      methodId = 0;
      final String componentCName = typeNameProviderItf.getCTypeName();
      cw.appendln("#include \"minus/bare/generic/exportbind/api/serialize.h\"");
      cw.appendln("#include \"minus/bare/generic/debug/api/debug.h\"");

      cw.append("#define REQUIRED ((struct ").append(componentCName).append(
          "_t *)_this)->type.imported ").endl();
      cw.append("#define NEWSTYLE_COMPONENT").endl();

      cw.append("  struct ").append(componentCName).append("_instancedata {};")
          .endl();
      cw.appendln("#include \"cecilia.h\"");
      cw.endl();
    }

    /**
     * Builds the footer part for a given interface definition.
     * 
     * @param itf the interface definition for which has to be built.
     * @throws Exception if any error occurs.
     */
    public void leaveInterface(final InterfaceDefinition itf) throws Exception {
      cw.append("#undef NEWSTYLE_COMPONENT").endl();
    }

    /**
     * Visits a field definition.
     * 
     * @param field the field definition to be used.
     * @throws Exception if any error occurs.
     */
    public void visitField(final Field field) throws Exception {
      // nothing to do
    }

    /**
     * Visits a method definition.
     * 
     * @param method the field definition to be used.
     * @throws Exception if any error occurs.
     */
    public void visitMethod(final Method method) throws Exception {
      final Type returnType = Util.getContainedType((method));
      final Parameter[] parameters = ((ParameterContainer) method)
          .getParameters();
      final boolean isVoidMethod = (returnType instanceof PrimitiveType)
          && getPrimitiveType((PrimitiveType) returnType) == PrimitiveTypeEnum.VOID;
      final Map<Parameter, Object> paramContext = new HashMap<Parameter, Object>();
      Object resultContext = null;
      boolean hasOutParam = false;

      // determine if there is an Out parameter
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p) || Util.isInOutParameter(p))
          hasOutParam = true;
      }

      final VarNameValidator localVars = new VarNameValidator();
      for (final Parameter p : parameters)
        localVars.validateName(p.getName());

      final String replyMsg = localVars.validateName("replyMsg");
      final String invokeMsg = localVars.validateName("invokeMsg");
      final String data = localVars.validateName("data");
      final String msgSize = localVars.validateName("msgSize");
      final String result = localVars.validateName("result");

      final String methodName = new StringBuffer().append(
          typeNameProviderItf.getCTypeName()).append("_itf_").append(
          method.getName()).append("_method").toString();

      // append method prototype
      cw.append(buildMethodDefinition(method, false, methodName)).append(" {")
          .endl();

      // declare invokeMsg and replyMsg
      cw.append("Rminus_bare_generic_exportbind_api_Message *").append(
          invokeMsg).append(";").endl();
      cw.append("Rminus_bare_generic_exportbind_api_Message *")
          .append(replyMsg).append(";").endl();

      // declare pointer to invokeMsg data to ease its manipulation
      cw.append("char *").append(data).append(";").endl();

      // declare local variable which will contain the return value
      if (!isVoidMethod) {
        cw.append(CUtil.buildReturnType(returnType)).append(" ").append(result)
            .append(";").endl();
      }

      // declaration related to parameter serialization
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p)) continue;
        final Object ctx = writeSerializationDeclaration(Util
            .getContainedType(p), p.getName(), cw, localVars);
        paramContext.put(p, ctx);
      }

      // declaration related to the result value deserialization
      if (!isVoidMethod) {
        resultContext = writeDeserializationDeclaration(returnType, result, cw,
            localVars);
      }

      // Compute invokeMsg size
      // The first byte will contain the method ID
      cw.append("int ").append(msgSize).append(" = 1;").endl();
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p)) continue;
        String paramName = p.getName();
        if (Util.isInOutParameter(p)) paramName = "(*" + paramName + ")";
        writeSizeOf(Util.getContainedType(p), paramName, msgSize, cw,
            paramContext.get(p));
      }

      // create invokeMsg
      cw.append(invokeMsg).append(" = ").append(
          "CALL(REQUIRED.message_manager, createMessage, ").append(msgSize)
          .append(");").endl();

      // initialize data pointer
      cw.append(data).append(" = (char *)invokeMsg->data;").endl().endl();

      // set method identifier
      cw.append("// Set method ID").endl();
      cw.append("*(").append(data).append("++) = ").append(methodId++).append(
          ";").endl().endl();

      // serialize parameters
      cw.append("// Serialize parameters").endl();
      for (final Parameter p : parameters) {
        if (Util.isOutParameter(p)) continue;
        cw.append("// Serialize \"" + p.getName() + "\"").endl();
        if (Util.isInOutParameter(p))
          writeSerialization(Util.getContainedType(p),
              "(*" + p.getName() + ")", data, cw, paramContext.get(p), false);
        else
          writeSerialization(Util.getContainedType(p), p.getName(), data, cw,
              paramContext.get(p), false);
      }
      cw.endl();

      // send invocation
      cw.append("// Send invokation").endl();
      cw.append("ASSERT(REQUIRED.invoke != 0);").endl();
      cw.append(replyMsg).append(" = CALL(REQUIRED.invoke, invoke, ").append(
          invokeMsg).append(");").endl().endl();

      // deserialize result value
      if (!isVoidMethod || hasOutParam) {
        cw.append("// Deserialize return value, out and in/out parameters")
            .endl();
        cw.append(data).append(" = (char *)replyMsg->data;").endl();
        cw.append("// Deserialize return value").endl();
        if (!isVoidMethod)
          writeDeserialization(returnType, result, data, cw, resultContext,
              true);
        for (final Parameter p : parameters) {
          if (Util.isOutParameter(p) || Util.isInOutParameter(p)) {
            cw.append("// Deserialize \"" + p.getName() + "\"").endl();
            writeDeserialization(Util.getContainedType(p), "(*" + p.getName()
                + ")", data, cw, resultContext, true);
          }
        }
        cw.endl();
        cw.append("CALL(REQUIRED.message_manager, deleteMessage, ").append(
            replyMsg).append(");").endl();
        if (!isVoidMethod)
          cw.appendln("return ").append(result).append(";").endl();
      } else {
        cw.append("CALL(REQUIRED.message_manager, deleteMessage, ").append(
            replyMsg).append(");").endl();
      }
      cw.appendln("}");
      cw.endl();
    }
  }
}
